#include "function_inline.hpp"
#include "../ir/ir.hpp"
#include "../parser/SyntaxTree.hpp"
#include "../passes/cfg.hpp"
#include <cassert>
#include <memory>
#include <unordered_map>
#include <unordered_set>

void Passes::FunctionInline::run() {
    bool changed = true;
    while(changed) {
        changed = false;
        for(auto [name, fun] : compunit->usrfuncs) {
            if(check_inlinable(fun, name)) {
                changed = true;
                inlinable.insert(fun);
                function_inline(fun);
            }
        }
        for(auto fun : inlinable) {
            compunit->remove_fun(fun);
        }
    }
    // if(inlinable.size()) {
    //     std::cout << "inlinable: ";
    //     for(auto fun : inlinable) {
    //         std::cout << fun->get_name() << " ";
    //     }
    //     std::cout << std::endl;
    // }
}

void Passes::FunctionInline::function_inline(ptr<ir::ir_userfunc> target) {
    std::unordered_set<ptr<ir::func_call>> call_set;
    for(auto weak_caller : target->get_caller()) {
        auto caller = weak_caller.lock();
        for(auto block : caller->get_bbs()) {
            for(auto ins : block->get_instructions()) {
                auto is_call = std::dynamic_pointer_cast<ir::func_call>(ins);
                if(is_call && is_call->get_callee() == target) {
                    call_set.insert(is_call);
                }
            }
        }
    }

    std::pmr::unordered_map<ptr<ir::ir_userfunc>, std::unordered_map<ptr<ir::ir_value>, ptr<ir::ir_value>>> arr_inserted;
    for(auto call : call_set) {
        auto ret_reg = call->get_ret_reg();
        auto block = call->get_block();
        auto position = block->search(call);
        auto caller = call->get_fun();
        std::unordered_map<ptr<ir::ir_value>, ptr<ir::ir_value>>& replace_map = arr_inserted[caller];
        if(replace_map.empty()) {
            for(auto obj : target->get_arrobj()) {
                caller->add_arrobj(obj);
            }
            for(auto alloc : target->get_var_list()) {
                auto cloned = std::dynamic_pointer_cast<ir::alloc>(alloc->clone());
                auto def = alloc->def_reg().front();
                auto new_reg = caller->new_reg(def->get_type());
                new_reg->clone_attribute(def);
                assert(replace_map.find(def) == replace_map.end());
                replace_map[def] = new_reg;
                alloc->replace_def(new_reg);
                caller->insert_alloc(alloc);
            }
        }
        // block->erase(position);

        auto end_block = caller->new_block();
        auto successor = caller->check_nxt(block);
        for(auto suc : successor) {
            for(auto ins : suc->get_instructions()) {
                auto is_phi = std::dynamic_pointer_cast<ir::phi>(ins);
                if(is_phi) {
                    std::unordered_map<ptr<ir::ir_basicblock>, std::weak_ptr<ir::ir_basicblock>> replace_map = {{block, end_block}};
                    is_phi->replace_block(replace_map);
                }
                else break;
            }
        }

        for(auto it = std::next(position); it != block->get_ins_end(); it++) {
            end_block->push_back(*it);
        }
        block->get_instructions_ref().erase(position, block->get_ins_end());

        auto entry = clone_function(call, target, end_block, ret_reg, replace_map);
        block->push_back(std::make_shared<ir::jump>(entry));

        CFGUtils::analyse_cfg(caller);
    }
}

ptr<ir::ir_basicblock> Passes::FunctionInline::clone_function(ptr<ir::func_call> call_ins, ptr<ir::ir_userfunc> target, ptr<ir::ir_basicblock> end_block, ptr<ir::ir_reg> inline_ret_reg, std::unordered_map<ptr<ir::ir_value>, ptr<ir::ir_value>> replace_map) {
    auto caller = call_ins->get_fun();
    auto entry = caller->new_block();
    auto cur_block = entry;
    std::unordered_map<ptr<ir::ir_basicblock>, ptr<ir::ir_basicblock>> visited;
    std::unordered_map<ptr<ir::ir_basicblock>, bool> phi_inserted;

    std::unordered_map<ptr<ir::ir_basicblock>, ptr<ir::ir_basicblock>> predecessor_map = {{entry, target->get_entry()}};     // cur_new_block - target
    visited[target->get_entry()] = entry;
    std::unordered_map<ptr<ir::ir_basicblock>, ptr<ir::ir_basicblock>> nxt_iter;     // cur_new_block - target
    std::unordered_map<ptr<ir::ir_basicblock>, ptr<ir::ir_basicblock>> phi_iter;     // cur_new_block - target
    // std::unordered_map<ptr<ir::ir_value>, ptr<ir::ir_value>> replace_map;
    auto params = target->get_params();
    auto call_params = call_ins->get_params();
    for(int i = 0; i < params.size(); i++) {
        auto par_reg = params[i]->get_addr();
        auto call_reg = call_params[i];
        replace_map[par_reg] = call_reg;
    }
    while(!predecessor_map.empty()) {
        nxt_iter.clear();
        phi_iter.clear();
        for(auto [empty_block, source_block] : predecessor_map) {

            auto instructions = source_block->get_instructions();
            for(auto ins = instructions.begin(); ins != instructions.end(); ins++) {
                auto is_phi = std::dynamic_pointer_cast<ir::phi>(*ins);
                if(is_phi)
                    continue;
                auto is_ret = std::dynamic_pointer_cast<ir::ret>(*ins);
                if(is_ret) {
                    if(target->get_rettype() != vartype::VOID) {
                        auto ret_reg = replace_map[is_ret->use_reg().front()];
                        assert(ret_reg->get_def_loc()->get_block() == visited[source_block]);
                        ret_reg->get_def_loc()->replace_def(inline_ret_reg);
                        inline_ret_reg->mark_def_loc(ret_reg->get_def_loc());
                    }
                    empty_block->push_back(std::make_shared<ir::jump>(end_block));
                    continue;
                }

                ptr<ir::ir_instr> cloned = ins->get()->clone();
                auto def_regs = (*ins)->def_reg();
                if(!def_regs.empty()) {
                    auto def = def_regs.front();
                    auto new_reg = caller->new_reg(def->get_type());
                    new_reg->clone_attribute(def);
                    assert(replace_map.find(def) == replace_map.end());
                    replace_map[def] = new_reg;
                    cloned->replace_def(new_reg);
                }
                cloned->replace_reg(replace_map);
                empty_block->push_back(cloned);
                if(ins == std::prev(instructions.end())) {
                    auto is_br = std::dynamic_pointer_cast<ir::br>(cloned);
                    auto is_jmp = std::dynamic_pointer_cast<ir::jump>(cloned);
                    auto is_while = std::dynamic_pointer_cast<ir::while_loop>(cloned);
                    auto is_bc = std::dynamic_pointer_cast<ir::break_or_continue>(cloned);
                    if(is_br) {
                        auto source_true = is_br->get_target_true();
                        auto source_false = is_br->get_target_false();
                        auto target_true = visited[source_true];
                        auto target_false = visited[source_false];
                        if(target_true == nullptr) {
                            target_true = caller->new_block();
                            nxt_iter[target_true] = source_true;
                            visited[source_true] = target_true;
                        }
                        if(target_false == nullptr) {
                            target_false = caller->new_block();
                            nxt_iter[target_false] = source_false;
                            visited[source_false] = target_false;
                        }
                        is_br->set_target_true(target_true);
                        is_br->set_target_false(target_false);
                        phi_iter[target_true] = source_true;
                        phi_iter[target_false] = source_false;
                    }
                    else if(is_jmp) {
                        auto source_tar = is_jmp->get_target();
                        auto tar = visited[source_tar];
                        if(tar == nullptr) {
                            tar = caller->new_block();
                            nxt_iter[tar] = source_tar;
                            visited[source_tar] = tar;
                        }
                        is_jmp->replace_target(tar);
                        phi_iter[tar] = source_tar;
                    }
                    else if(is_while) {
                        auto source_cond_from = is_while->get_cond_from();
                        auto source_out_block = is_while->get_out_block();
                        auto cond_from = visited[source_cond_from];
                        auto out_block = visited[source_out_block];
                        if(cond_from == nullptr) {
                            cond_from = caller->new_block();
                            nxt_iter[cond_from] = source_cond_from;
                            visited[source_cond_from] = cond_from;
                        }
                        if(out_block == nullptr) {
                            out_block = caller->new_block();
                            nxt_iter[out_block] = source_out_block;
                            visited[source_out_block] = out_block;
                        }
                        is_while->set_cond_from(cond_from);
                        is_while->set_out_block(out_block);
                        phi_iter[cond_from] = source_cond_from;
                        phi_iter[out_block] = source_out_block;
                    }
                    else if(is_bc) {
                        auto source_tar = is_bc->get_target();
                        auto tar = visited[source_tar];
                        if(tar == nullptr) {
                            tar = caller->new_block();
                            nxt_iter[tar] = source_tar;
                            visited[source_tar] = tar;
                        }
                        is_bc->set_target(tar);
                        phi_iter[tar] = source_tar;
                    }
                    else {
                        abort();
                    }
                    for(auto [phi_empty_block, phi_source_block] : phi_iter) {
                        if(!phi_inserted[phi_empty_block]) {
                            for(auto ins : phi_source_block->get_instructions()) {
                                auto is_phi = std::dynamic_pointer_cast<ir::phi>(ins);
                                if(is_phi) {
                                    auto phi_ins = std::dynamic_pointer_cast<ir::phi>(is_phi->clone());
                                    auto def = phi_ins->def_reg().front();
                                    auto new_reg = caller->new_reg(def->get_type());
                                    new_reg->clone_attribute(def);
                                    assert(replace_map.find(def) == replace_map.end());
                                    replace_map[def] = new_reg;
                                    phi_ins->replace_def(new_reg);
                                    // phi_ins->replace_reg(replace_map);
                                    // std::unordered_map<ptr<ir::ir_basicblock>, std::weak_ptr<ir::ir_basicblock>> replace_block_map = {{source_block, empty_block}};
                                    // phi_ins->replace_block(replace_block_map);
                                    phi_ins->set_rank(is_phi->get_rank());
                                    phi_empty_block->push_back(phi_ins);
                                    phi_inserted[phi_empty_block] = true;
                                }
                                else
                                    break;
                            }
                        }
                        for(auto ins : phi_empty_block->get_instructions()) {
                            auto is_phi = std::dynamic_pointer_cast<ir::phi>(ins);
                            if(is_phi) {
                                // auto def = is_phi->def_reg().front();
                                // auto new_reg = caller->new_reg(def->get_type());
                                // new_reg->clone_attribute(def);
                                // assert(replace_map.find(def) == replace_map.end());
                                // replace_map[def] = new_reg;
                                // is_phi->replace_def(new_reg);
                                is_phi->replace_reg(replace_map);
                                std::unordered_map<ptr<ir::ir_basicblock>, std::weak_ptr<ir::ir_basicblock>> replace_block_map = {{source_block, empty_block}};
                                is_phi->replace_block(replace_block_map);
                            }
                            else
                                break;
                        }
                    }
                }
            }

        }
        predecessor_map = nxt_iter;
    }
    
    return entry;
}

void Passes::FunctionInline::clone_ins(ptr<ir::ir_userfunc> caller, std::list<std::shared_ptr<ir::ir_instr>> instructions) {
    std::unordered_map<ptr<ir::ir_value>, ptr<ir::ir_value>> replace_map;
    for(auto ins : instructions) {
        // ptr<ir::ir_instr> cloned = ins.get()->clone();
        auto def = ins->def_reg().front();
        auto new_reg = caller->new_reg(def->get_type());
        assert(replace_map.find(def) == replace_map.end());
        replace_map[def] = new_reg;
        // cloned->replace_reg(replace_map);
    }
}

bool Passes::FunctionInline::check_inlinable(ptr<ir::ir_userfunc> fun, std::string name) {
    if(name == "main" || name == "global_init") return false;
    int size = 0;
    for(auto block : fun->get_bbs()) {
        size += block->get_instructions().size();
    }
    if(size >= INLINE_SIZE_LIMIT)
        return false;
    for(auto block : fun->get_bbs()) {
        for(auto ins : block->get_instructions()) {
            auto is_call = std::dynamic_pointer_cast<ir::func_call>(ins);
            // TODO: Tail call 如必要需要删掉
            if(is_call) {
                auto callee = std::dynamic_pointer_cast<ir::ir_userfunc>(is_call->get_callee());
                if(callee && callee->get_name() == name) {
                    return false;
                }
            }
        }
    }
    return inlinable.find(fun) == inlinable.end();
}